/**
 * APBUART driver for RTEMS.
 * Uses the Termios framework in interrupt-driven mode.
 *
 * To use this driver, either call leonuart_register() during initialization
 * initialization of your application, or add LEONUART_DRIVER_TABLE_ENTRY
 * to the RTEMS device driver table.
 *
 * Device nodes are registered as /dev/leonuart0 ...
 * Baud rate can be changed through tcsetattr().
 * The following driver-specific ioctls are also available:
 * 
 * Set non-standard baud rate:
 *   unsigned int bps;
 *   ioctl(fd, LEONUART_IOC_SET_BAUDRATE, &bps);
 *
 * Clear error/byte counters:  
 *   ioctl(fd, LEONUART_IOC_CLR_STATS, 0);
 *
 * Read error/byte counters:
 *   struct leonuart_stats stats;
 *   ioctl(fd, LEONUART_IOC_GET_STATS, &stats);
 *
 * The driver supports APBUARTs with either FIFOs or holding registers.
 * At 115200 bps, it may be necessary to increase the Termios receive buffer
 * size through rtems_termios_bufsize().
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>

#include <rtems.h>
#include <rtems/libio.h>
#include <rtems/termiostypes.h>
#include <bsp.h>
#include <rtems/bspIo.h>
#include <ambapp.h>

#include "leonuart.h"


/* UART register bits */
#define APBUART_CTRL_TF     0x200
#define APBUART_CTRL_FA     0x80000000
#define APBUART_STATUS_TF   0x200

/* The LEON3 has AMBA plug&play data for device identification. */
#ifdef LEON3

#define GET_SYSFREQ_MHZ() \
    ( LEON3_Timer_Regs ? (LEON3_Timer_Regs->scaler_reload + 1) : 0 )

#define GET_N_BUILTIN_UARTS() \
    ( amba_get_number_apbslv_devices(&amba_conf, VENDOR_GAISLER, GAISLER_APBUART) )

#define GET_BUILTIN_UART(i, apbdev) \
    ( amba_find_next_apbslv(&amba_conf, VENDOR_GAISLER, GAISLER_APBUART, &(apbdev), (i)) )

#endif

/* The LEON2 has a fixed address map. */
#ifdef LEON2

#define GET_SYSFREQ_MHZ()           ( LEON_REG.Scaler_Reload + 1 )

#define GET_N_BUILTIN_UARTS()       ( 2 )

#define GET_BUILTIN_UART(i, apbdev) \
    ( (apbdev).start = 0x80000070 + ((i) << 4), (apbdev).irq = 3 - (i) )

#endif


/* Per-device variables */
struct leonuart_dev {
    struct rtems_termios_tty *tty;
    ambapp_apb_uart *regs;
    unsigned int    irq;
    unsigned int    txpending;
    unsigned int    baud_override;
    struct leonuart_stats stats;
};


static unsigned int leonuart_ndevs = 0;
static unsigned int leonuart_sysfreq_mhz = 0;
static struct leonuart_dev *leonuart_devtable = NULL;

static unsigned int leonuart_n_extra_uarts = 0;
static const amba_apb_device *leonuart_extra_uarts = NULL;


/* Convert baud rate to APBUART scaler reload */
static unsigned int baud_to_scaler(unsigned int baud)
{
    unsigned int scaler;
    if (baud == 0) baud = 1;
    scaler = ((leonuart_sysfreq_mhz * 125000 * 10) / baud - 5) / 10;
    if (scaler > 0xfff) scaler = 0xfff;
    return scaler;
}


/* ==========  Termios interface  ========== */


static void leonuart_interrupt_handler(rtems_vector_number vector)
{
    char buf[16];
    unsigned int i;

    /* Find device by vector number */
    for (i = 0; i < leonuart_ndevs; i++) {
        struct leonuart_dev *dev = leonuart_devtable + i;
        unsigned int ctrl, status, k;

        /* Is the right APBUART for this IRQ ? */
        if (0x10 + dev->irq != vector)
            continue;

        /* Is this APBUART linked to a tty ? */
        if (!dev->tty)
            continue;

        /* Check for errors */
        status = dev->regs->status;
        if (status & LEON_REG_UART_STATUS_ERR) {
            if (status & LEON_REG_UART_STATUS_BR) dev->stats.brk++;
            if (status & LEON_REG_UART_STATUS_OE) dev->stats.overrun++;
            if (status & LEON_REG_UART_STATUS_PE) dev->stats.parity++;
            if (status & LEON_REG_UART_STATUS_FE) dev->stats.frame++;
            dev->regs->status = 0;
        }

        /* Check for received data */
        for (k = 0; dev->regs->status & LEON_REG_UART_STATUS_DR; k++) {
            if (k == sizeof(buf)) {
                dev->stats.rx += k;
                rtems_termios_enqueue_raw_characters(dev->tty, buf, k);
                k = 0;
            }
            buf[k] = (unsigned char)dev->regs->data;
        }
        if (k) {
            dev->stats.rx += k;
            rtems_termios_enqueue_raw_characters(dev->tty, buf, k);
        }

        /* Check for transmitted data */
        ctrl   = dev->regs->ctrl;
        status = dev->regs->status;
	if ((status & LEON_REG_UART_STATUS_THE) != 0 ||
            ((ctrl & APBUART_CTRL_FA) != 0 &&
             (status & APBUART_STATUS_TF) == 0)) {
            /* There is room in the FIFO; turn off TX interrupts */
            dev->regs->ctrl = ctrl & ~(LEON_REG_UART_CTRL_TI | APBUART_CTRL_TF);
            /* Acknowledge pending characters */
            if (dev->txpending) {
                k = dev->txpending;
                dev->txpending = 0;
                rtems_termios_dequeue_characters(dev->tty, k);
            }
        }
    }
}


static int leonuart_interrupt_write(int minor, const char *buf, int n)
{
    struct leonuart_dev *dev = leonuart_devtable + minor;
    unsigned int ctrl;
    int k;

    /* Termios calls this function with interrupts disabled ... */

    /* Enable TX interrupts.
       The interrupt is event-triggered so it must be enabled before
       writing data to the UART to avoid potentially missing the event. */
    ctrl = dev->regs->ctrl;
    dev->regs->ctrl = ctrl | LEON_REG_UART_CTRL_TI;

    if (ctrl & APBUART_CTRL_FA) {
        /* The UART has a FIFO; we can send multiple bytes */
        for (k = 0; k < n && (dev->regs->status & APBUART_STATUS_TF) == 0; k++)
            dev->regs->data = (unsigned char)buf[k];
        /* Additionally enable FIFO transmit interrupt.
           This interrupt is level-triggered so it's ok to enable it late. */
        dev->regs->ctrl = ctrl | LEON_REG_UART_CTRL_TI | APBUART_CTRL_TF;
    } else {
        /* No FIFO; only holding register */
        for (k = 0; k < n && (dev->regs->status & LEON_REG_UART_STATUS_THE) != 0; k++)
            dev->regs->data = (unsigned char)buf[k];
    }

    dev->txpending += k;
    dev->stats.tx  += k;

    return 0;
}


static int leonuart_first_open(int major, int minor, void *arg)
{
    struct leonuart_dev *dev;
    int i;

    /* Check minor number */
    if (minor < 0 || (unsigned int)minor >= leonuart_ndevs)
        return -1;

    dev = leonuart_devtable + minor;

    dev->txpending = 0;
    dev->baud_override = 0;

    /* Connect the TTY data structure */
    dev->tty = (struct rtems_termios_tty *)((rtems_libio_open_close_args_t *)arg)->iop->data1;

    /* Register interrupt. */
    set_vector(leonuart_interrupt_handler, 0x10 + dev->irq, 1);

    /* Initialize device and flush RX buffer. */
    dev->regs->ctrl = 0;
    dev->regs->status = 0;
    dev->regs->scaler = baud_to_scaler(38400);
    for (i = 0; i < 16 && (dev->regs->status & LEON_REG_UART_STATUS_DR) != 0;
         i++)
        dev->regs->data;
    dev->regs->ctrl = LEON_REG_UART_CTRL_RE | LEON_REG_UART_CTRL_TE |
                      LEON_REG_UART_CTRL_RI ;

    /* Set the initial baud rate */
    return rtems_termios_set_initial_baud(dev->tty, 38400);
}


static int leonuart_last_close(int major, int minor, void *arg)
{
    struct leonuart_dev *dev = leonuart_devtable + minor;

    /* Shut down the device. */
    dev->regs->ctrl = 0;

    /* Disconnect the TTY data structure */
    dev->tty = NULL;

    return 0;
}


static int leonuart_set_attributes(int minor, const struct termios *t)
{
    struct leonuart_dev *dev = leonuart_devtable + minor;
    rtems_interrupt_level level;
    unsigned int ctrl, scaler;

    /* Compute new baud rate scaler. */
    scaler = baud_to_scaler(rtems_termios_baud_to_number(cfgetospeed(t)));

    /* Disable interrupts to avoid interference from our int handler. */
    rtems_interrupt_disable(level);

    /* Set parity mode and flow control */
    ctrl = dev->regs->ctrl;
    ctrl &= ~( LEON_REG_UART_CTRL_PE | LEON_REG_UART_CTRL_PS |
               LEON_REG_UART_CTRL_FL );
    if (t->c_cflag & PARENB)  ctrl |= LEON_REG_UART_CTRL_PE;
    if (t->c_cflag & PARODD)  ctrl |= LEON_REG_UART_CTRL_PS;
    if (t->c_cflag & CRTSCTS) ctrl |= LEON_REG_UART_CTRL_FL;
    dev->regs->ctrl = ctrl;

    /* Set baud rate */
    if (!dev->baud_override)
        dev->regs->scaler = scaler;

    rtems_interrupt_enable(level);

    return 0;
}


/* Termios callbacks */
static const rtems_termios_callbacks leonuart_callbacks = {
    leonuart_first_open,        /* firstOpen */
    leonuart_last_close,        /* lastClose */
    NULL,                       /* pollRead */
    leonuart_interrupt_write,   /* write */
    leonuart_set_attributes,    /* setAttributes */
    NULL,                       /* stopRemoteTx */
    NULL,                       /* startRemoteTx */
    TERMIOS_IRQ_DRIVEN          /* outputUsesInterrupts */
};


/* ==========  Device driver interface  ========== */


rtems_device_driver leonuart_initialize(rtems_device_major_number major,
                                        rtems_device_minor_number minor,
                                        void *arg)
{
    rtems_device_minor_number i;
    unsigned int n_builtin_uarts;

    /* Initialize Termios framework, if not done before. */
    rtems_termios_initialize();

    /* Detect system clock frequency for baud rate computations. */
    leonuart_sysfreq_mhz = GET_SYSFREQ_MHZ();
    if (!leonuart_sysfreq_mhz) {
        printk("leonuart: Can not detect system frequency\n");
        return RTEMS_NOT_CONFIGURED;
    }
    printk("leonuart: Assuming system frequency %u MHz\n", leonuart_sysfreq_mhz);

    /* Count number of UARTs. */
    n_builtin_uarts = GET_N_BUILTIN_UARTS();
    leonuart_ndevs = n_builtin_uarts + leonuart_n_extra_uarts;

    /* Allocate memory for device structures. */
    if (leonuart_ndevs < 1) {
        printk("leonuart: No APBUART cores found\n");
    } else {
        leonuart_devtable = (struct leonuart_dev *)malloc(sizeof(struct leonuart_dev) * leonuart_ndevs);
        if (!leonuart_devtable) {
            printk("leonuart: ERROR: Can not allocate memory\n");
            return RTEMS_NO_MEMORY;
        }
    }

    /* Initialize device structures. */
    for (i = 0; i < leonuart_ndevs; i++) {
        amba_apb_device apbdev;
        if (i < n_builtin_uarts) {
            GET_BUILTIN_UART(i, apbdev);
        } else {
            apbdev = leonuart_extra_uarts[i-n_builtin_uarts];
        }
        leonuart_devtable[i].tty   = NULL;
        leonuart_devtable[i].regs  = (ambapp_apb_uart *)apbdev.start;
        leonuart_devtable[i].irq   = apbdev.irq;
        leonuart_devtable[i].txpending = 0;
        leonuart_devtable[i].baud_override = 0;
        memset(&leonuart_devtable[i].stats, 0, sizeof(leonuart_devtable[i].stats));
    }

    /* Register devices. */
    for (i = 0; i < leonuart_ndevs; i++) {
        char devname[20];
        sprintf(devname, "/dev/leonuart%d", (int)i);
	if (rtems_io_register_name(devname, major, i) != RTEMS_SUCCESSFUL)
            printk("leonuart: ERROR: Could not register %s\n", devname);
        else
            printk("leonuart: Registered %s at 0x%08x irq %u\n", devname, (unsigned int)leonuart_devtable[i].regs, leonuart_devtable[i].irq);
    }

    return RTEMS_SUCCESSFUL;
}


rtems_device_driver leonuart_open(rtems_device_major_number major,
                                  rtems_device_minor_number minor,
                                  void *arg)
{
    /* Check minor number */
    if (minor < 0 || minor >= leonuart_ndevs)
        return RTEMS_INVALID_NUMBER;

    return rtems_termios_open(major, minor, arg, &leonuart_callbacks);
}


rtems_device_driver leonuart_close( rtems_device_major_number major,
                                    rtems_device_minor_number minor,
                                    void *arg)
{
    return rtems_termios_close(arg);
}


rtems_device_driver leonuart_read(rtems_device_major_number major,
                                  rtems_device_minor_number minor,
                                  void *arg)
{
    return rtems_termios_read(arg);
}


rtems_device_driver leonuart_write( rtems_device_major_number major,
                                    rtems_device_minor_number minor,
                                    void *arg)
{
    return rtems_termios_write(arg);
}


rtems_device_driver leonuart_control(rtems_device_major_number major,
                                     rtems_device_minor_number minor,
                                     void *arg)
{
    struct leonuart_dev *dev = leonuart_devtable + minor;
    rtems_libio_ioctl_args_t *ioarg = (rtems_libio_ioctl_args_t *)arg;
    unsigned int baudrate;

    if (!ioarg)
        return RTEMS_INVALID_NAME;

    /* Handle non-Termios ioctls */
    switch (ioarg->command) {

        case LEONUART_IOC_SET_BAUDRATE:
            /* Set non-standard baud rate */
            if (!ioarg->buffer)
                return RTEMS_INVALID_NAME;
            baudrate = *(unsigned int *)ioarg->buffer;
            if (baudrate < 1 || baudrate > leonuart_sysfreq_mhz * 125000)
                return RTEMS_INVALID_NAME;
            dev->baud_override = baudrate;
            if (dev->baud_override)
                dev->regs->scaler = baud_to_scaler(baudrate);
            return RTEMS_SUCCESSFUL;
	
        case LEONUART_IOC_GET_STATS:
            /* Get device statistics */
            if (!ioarg->buffer)
                return RTEMS_INVALID_NAME;
            *(struct leonuart_stats *)ioarg->buffer = dev->stats;
            return RTEMS_SUCCESSFUL;
  
        case LEONUART_IOC_CLR_STATS:
            /* Reset device statistics */
            memset(&dev->stats, 0, sizeof(dev->stats));
            return RTEMS_SUCCESSFUL;
    }

    /* Handle Termios ioctls */
    return rtems_termios_ioctl(arg);
}


static const rtems_driver_address_table leonuart_driver = {
    leonuart_initialize,        /* initialization_entry */
    leonuart_open,              /* open_entry */
    leonuart_close,             /* close_entry */
    leonuart_read,              /* read_entry */
    leonuart_write,             /* write_entry */
    leonuart_control            /* control_entry */
};


int leonuart_register(void)
{
    rtems_device_major_number major;
    return rtems_io_register_driver(0, &leonuart_driver, &major);
}


void leonuart_set_extra_uarts(const amba_apb_device *uarts, unsigned int n)
{
    leonuart_n_extra_uarts = n;
    leonuart_extra_uarts = uarts;
}

/* end */
