/**
 * Functional/performance test of serial port.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <apbuart.h>
#include <bsp.h>
#include <rtems.h>
#include <rtems/bspIo.h>
#include <rtems/rtems_bsdnet.h>

#include "leonuart.h"

int main(void);

#define CONFIGURE_INIT
#define CONFIGURE_APPLICATION_NEEDS_CONSOLE_DRIVER
#define CONFIGURE_APPLICATION_NEEDS_CLOCK_DRIVER
#define CONFIGURE_APPLICATION_NEEDS_TIMER_DRIVER
#define CONFIGURE_RTEMS_INIT_TASKS_TABLE
#define CONFIGURE_INIT_TASK_ENTRY_POINT main
#define CONFIGURE_NUMBER_OF_TERMIOS_PORTS  4
#define CONFIGURE_LIBIO_MAXIMUM_FILE_DESCRIPTORS 128
#define CONFIGURE_MAXIMUM_TIMERS  16
#define CONFIGURE_MAXIMUM_TASKS   16
#define CONFIGURE_MAXIMUM_DRIVERS 16
#define CONFIGURE_MINIMUM_TASK_STACK_SIZE  16384
#define CONFIGURE_INIT_TASK_PRIORITY 200
#include <rtems/confdefs.h>


/* Loopback interface */
int rtems_bsdnet_loopattach(struct rtems_bsdnet_ifconfig *, int);
static struct rtems_bsdnet_ifconfig loopback_config = {
    "lo0",                          /* name */
    rtems_bsdnet_loopattach,        /* attach function */
    NULL,                           /* link to next interface */
    "127.0.0.1",                    /* ip_address */
    "255.0.0.0",                    /* ip_netmask */
};

/* Ethernet */
static struct rtems_bsdnet_ifconfig netdriver_config = {
    "gr_eth1",                      /* name */
    rtems_leon_greth_driver_attach, /* attach */
    &loopback_config,               /* link to next interface */
    "192.168.0.80",                 /* ip_address */
    "255.255.255.0",                /* ip_netmask */
};

/* Network configuration */
struct rtems_bsdnet_config rtems_bsdnet_config = {
    &netdriver_config,              /* ifconfig */
    NULL,                           /* bootp */
    100,                            /* network_task_priority */
    128*1024,                       /* mbuf_bytecount */
    256*1024,                       /* mbuf_cluster_bytecount */
    "rtems",                        /* hostname */
    "localnet",                     /* domainname */
    "192.168.0.1",                  /* gateway */
    NULL,                           /* log host */
    {NULL, },                       /* Name server(s) */
    {NULL, },                       /* NTP server(s) */
};

static char buf[1024];
static int  netsock;

static void output_char(char c)
{
    write(netsock, &c, 1);
}

static void netprintf(const char *format, ...)
{
    int i, n;
    va_list ap;
    va_start(ap, format);
    vsnprintf(buf, sizeof(buf), format, ap);
    n = strlen(buf);
    for (i = 0; i < n; ) {
        int k = write(netsock, buf + i, n - i);
        if (k > 0) i += k;
    }
    va_end(ap);
}

/* Read from 32-bit register, bypassing MMU and cache. */
static inline unsigned int readreg32(unsigned int addr)
{
    unsigned int v;
    asm volatile (
        " lda [%1] 1, %0 "
        : "=r" (v)
        : "r" (addr) );
    return v;
}

/* Write to 32-bit register, bypassing MMU and cache. */
static inline void writereg32(unsigned int addr, unsigned int v)
{
    asm volatile (
        " sta %0, [%1] 1 "
        : : "r" (v), "r" (addr)
        : "memory" );
}

static void try_readwrite(int fd)
{
    char buf[64];
    struct timespec tstart, tend;
    long totwrite = 0, totread = 0, drop = 0;
    int i;

    /* flush read buf */
    for (i = 0; i < 8; i++)
        read(fd, buf, sizeof(buf));

    /* send data for 5 seconds */
    clock_gettime(CLOCK_REALTIME, &tstart);
    do {
        int k;
        for (i = 0; i < sizeof(buf); i++)
            buf[i] = (unsigned char)(totwrite + i);
        k = write(fd, buf, sizeof(buf) / 2);
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error write\n");
            return;
        }
        if (k > 0)
            totwrite += k;
        k = read(fd, buf, sizeof(buf));
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error read\n");
            return;
        }
        for (i = 0; i < k; i++)
            if ((unsigned char)buf[i] != (unsigned char)(totread + i))
                drop = 1;
        if (k > 0)
            totread += k;
        clock_gettime(CLOCK_REALTIME, &tend);
    } while (tend.tv_sec < tstart.tv_sec + 5 ||
             (tend.tv_sec == tstart.tv_sec + 5 && tend.tv_nsec < tstart.tv_nsec));
    /* flush read buf */
    read(fd, buf, sizeof(buf));

    /* report speed */
    netprintf("  readwrite: tx %ld / rx %ld bytes (%s)\n", totwrite, totread, drop ? "DATA DROPPED" : "ok");
}

static void try_read(int fd)
{
    char buf[64], p;
    struct timespec tstart, tend;
    long totread = 0, drop = 0;
    int i;

    /* flush read buf */
    for (i = 0; i < 8; i++)
        read(fd, buf, sizeof(buf));

    /* read data for 5 seconds */
    clock_gettime(CLOCK_REALTIME, &tstart);
    p = 0;
    do {
        int k;
        k = read(fd, buf, sizeof(buf));
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error read\n");
            return;
        }
        if (totread == 0)
            p = buf[0];
        for (i = 0; i < k; i++)
            if (buf[i] != p++)
                drop = 1;
        if (k > 0)
            totread += k;
        clock_gettime(CLOCK_REALTIME, &tend);
    } while (tend.tv_sec < tstart.tv_sec + 5 ||
             (tend.tv_sec == tstart.tv_sec + 5 && tend.tv_nsec < tstart.tv_nsec));

    /* report speed */
    netprintf("  read: rx %ld bytes (%s)\n", totread, drop ? "DATA DROPPED" : "ok");
}

static void try_write(int fd)
{
    char buf[64];
    struct timespec tstart, tend;
    long totwrite = 0;

    /* send data for 5 seconds */
    clock_gettime(CLOCK_REALTIME, &tstart);
    do {
        int i, k;
        for (i = 0; i < sizeof(buf); i++)
            buf[i] = (unsigned char)(totwrite + i);
        k = write(fd, buf, sizeof(buf));
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error write\n");
            return;
        }
        if (k > 0)
            totwrite += k;
        clock_gettime(CLOCK_REALTIME, &tend);
    } while (tend.tv_sec < tstart.tv_sec + 5 ||
             (tend.tv_sec == tstart.tv_sec + 5 && tend.tv_nsec < tstart.tv_nsec));

    /* report speed */
    netprintf("  write: tx %ld bytes\n", totwrite);
}

static void try_console(const char *devname, int fdrx, int fdtx)
{
    struct termios attr;
    char buf[8];
    struct timespec tstart, tend;
    long totwrite = 0, totread = 0, drop = 0;
    int i;

    netprintf("  %s:", devname);

    /* configure console */
    attr.c_iflag = IGNBRK | IGNPAR;
    attr.c_oflag = 0;
    attr.c_cflag = CS8 | CREAD | CLOCAL;
    attr.c_lflag = 0;
    memset(attr.c_cc, 0, sizeof(attr.c_cc));
    attr.c_cc[VMIN] = 0;
    attr.c_cc[VTIME] = 0;
    cfsetispeed(&attr, B38400);
    cfsetospeed(&attr, B38400);
    if (tcsetattr(fdtx, TCSADRAIN, &attr) < 0) {
	netprintf("error tcsetattr\n");
        return;
    }

    /* flush read buf */
    while (read(fdrx, buf, sizeof(buf)) > 0) ;

    /* send data for 1 second */
    clock_gettime(CLOCK_REALTIME, &tstart);
    do {
        int k;
	for (i = 0; i < sizeof(buf); i++)
            buf[i] = (unsigned char)(totwrite + i);
        k = write(fdtx, buf, sizeof(buf));
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error write\n");
            return;
        }
        if (k > 0)
            totwrite += k;
        k = read(fdrx, buf, sizeof(buf));
        if (k < 0 && errno != EAGAIN && errno != ETIMEDOUT) {
            netprintf("error read\n");
            return;
        }
        for (i = 0; i < k; i++)
            if ((unsigned char)buf[i] != (unsigned char)(totread + i))
                drop = 1;
        if (k > 0)
            totread += k;
        clock_gettime(CLOCK_REALTIME, &tend);
    } while (tend.tv_sec < tstart.tv_sec + 5 ||
             (tend.tv_sec == tstart.tv_sec + 5 && tend.tv_nsec < tstart.tv_nsec));

    /* flush read buf */
    while (read(fdrx, buf, sizeof(buf)) > 0) ;

    /* report speed */
    netprintf("  tx %ld / rx %ld bytes (%s)\n", totwrite, totread, drop ? "DATA DROPPED" : "ok");
}

static void try_leonuart(const char *devname, unsigned int baud)
{
    struct termios attr;
    struct leonuart_stats stats;
    int fd;

    netprintf("  %s (baud=%d):\n", devname, baud);

    /* open port */
    fd = open(devname, O_RDWR);
    if (fd < 0) {
        netprintf("error open\n");
        return;
    }

    /* configure port */
    attr.c_iflag = IGNBRK | IGNPAR;
    attr.c_oflag = 0;
    attr.c_cflag = CS8 | CREAD | CLOCAL;
    attr.c_lflag = 0;
    memset(attr.c_cc, 0, sizeof(attr.c_cc));
    attr.c_cc[VMIN] = 0;    /* non-blocking */
    attr.c_cc[VTIME] = 0;
    cfsetispeed(&attr, B115200);
    cfsetospeed(&attr, B115200);
    if (tcsetattr(fd, TCSADRAIN, &attr) < 0) {
	netprintf("error tcsetattr\n");
        return;
    }
    if (baud != 115200)
        ioctl(fd, LEONUART_IOC_SET_BAUDRATE, &baud);

    if (ioctl(fd, LEONUART_IOC_CLR_STATS, 0) < 0)
        netprintf("error ioctl(LEONUART_CLR_STATS)\n");

    try_readwrite(fd);

    /* read error counts */
    if (ioctl(fd, LEONUART_IOC_GET_STATS, &stats) < 0) {
        netprintf("error ioctl(LEONUART_GET_STATS)\n");
    } else {
        netprintf("  stats: overrun=%u parity=%u frame=%u break=%u rx=%u tx=%u\n", stats.overrun, stats.parity, stats.frame, stats.brk, stats.rx, stats.tx);
    }
    sleep(1);

    try_read(fd);
    try_write(fd);

    close(fd);
}

static void try_apbuart(const char *devname, unsigned int scaler)
{
    apbuart_stats stats;
    int fd;

    netprintf("  %s (scaler=%d,baud=%d):\n",
           devname, scaler, 40000000 / 8 / (scaler + 1));

    fd = open(devname, O_RDWR);
    if (fd < 0) {
        perror("open");
        return;
    }

    /* setup serial port */
    if (ioctl(fd, APBUART_SET_RXFIFO_LEN, 512) < 0 ||
        ioctl(fd, APBUART_SET_TXFIFO_LEN, 512) < 0) {
        netprintf("error ioctl(APBUART_SET_??FIFO_LEN)\n");
        return;
    }
    if (ioctl(fd, APBUART_SET_SCALER, &scaler) < 0) {
        netprintf("error ioctl(APBUART_SET_SCALER)\n");
        return;
    }
    if (ioctl(fd, APBUART_SET_BLOCKING, 0) < 0) {
        netprintf("error ioctl(APBUART_SET_BLOCKING)\n");
        return;
    }
    if (ioctl(fd, APBUART_CLR_STATS, 0) < 0) {
        netprintf("error ioctl(APBUART_CLR_STATS)\n");
    }
    if (ioctl(fd, APBUART_START, NULL) < 0) {
        netprintf("error ioctl(APBUART_START)\n");
        return;
    }

    try_readwrite(fd);

    /* read error counts */
    if (ioctl(fd, APBUART_GET_STATS, &stats) < 0) {
        netprintf("error ioctl(APBUART_GET_STATS)\n");
    } else {
        netprintf("  stats: hw_dovr=%u parity=%u frame=%u sw_dovr=%u rx_cnt=%u tx_cnt=%u\n", stats.hw_dovr, stats.hw_parity, stats.hw_frame, stats.sw_dovr, stats.rx_cnt, stats.tx_cnt);
    }

    try_read(fd);
    try_write(fd);

    /* stop serial port */
    if (ioctl(fd, APBUART_STOP, NULL) < 0) {
        netprintf("error ioctl(APBUART_STOP)\n");
        return;
    }

    close(fd);
}

int main(void)
{
    int s;
    struct sockaddr_in addr;
    socklen_t addrlen;

    rtems_bsdnet_initialize_network();
    s = socket(PF_INET, SOCK_STREAM, 0);
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(5011);
    bind(s, (struct sockaddr *)&addr, sizeof(addr));
    listen(s, 1);
    addrlen = sizeof(addr);
    netsock = accept(s, (struct sockaddr *)&addr, &addrlen);
    close(s);
    netprintf("Hello kitty\n");
    sleep(1);

    BSP_output_char = output_char;

    amba_print_conf(&amba_conf);

#if 0
    apbuart_register(&amba_conf);
    sleep(1);

    try_apbuart("/dev/apbuart1", 42);
    sleep(1);
#endif

#if 1
    rtems_termios_bufsize(512, 512, 512);
    leonuart_register();
    sleep(1);

    try_leonuart("/dev/leonuart1", 115200);
#endif

    netprintf(".\n");

    pause();
    return 0;
}

